/*****************************************************************************
  Copyright (C) 2024 by Björn Lemke
  
  This code is derived from MariaDB/Connector written by Georg Richter
 
  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not see <http://www.gnu.org/licenses>
  or write to the Free Software Foundation, Inc.,
  51 Franklin St., Fifth Floor, Boston, MA 02110, USA
 *****************************************************************************/

#include "cego_python.h"

char *dsn_keys[]= {
    "dsn", "host", "port", "protocol", "tableset", "user", "password", "pool_name", "pool_size",
    NULL
};

static void
CegoConnection_finalize(CegoConnection* self);

static PyObject*
CegoConnection_exception(PyObject* self, void* closure);

#define GETTER_EXCEPTION(name, exception, doc)\
{ name,CegoConnection_exception, NULL, doc, &exception }


static PyObject* CegoConnection_connection_id(CegoConnection* self);

static PyObject* CegoConnection_getinfo(CegoConnection* self, PyObject* optionval);

static PyObject* CegoConnection_executecommand(CegoConnection* self, PyObject* command);

PyObject* CegoConnection_connect( PyObject* self, PyObject* args, PyObject* kwargs);

static PyObject* CegoConnection_close(CegoConnection* self);

static PyObject* CegoConnection_ping(CegoConnection* self);

static PyGetSetDef
CegoConnection_sets[]=
{
    {"connection_id", (getter)CegoConnection_connection_id,
        NULL, "Id of current connection", NULL},
    GETTER_EXCEPTION("Error", Cego_Error, ""),
    GETTER_EXCEPTION("Warning", Cego_Warning, exception_warning__doc__),
    GETTER_EXCEPTION("InterfaceError", Cego_InterfaceError, exception_interface__doc__),
    GETTER_EXCEPTION("ProgrammingError", Cego_ProgrammingError, exception_programming__doc__),
    GETTER_EXCEPTION("IntegrityError", Cego_IntegrityError, exception_integrity__doc__),
    GETTER_EXCEPTION("DatabaseError", Cego_DatabaseError, exception_database__doc__),
    GETTER_EXCEPTION("NotSupportedError", Cego_NotSupportedError, exception_notsupported__doc__),
    GETTER_EXCEPTION("InternalError", Cego_InternalError, exception_internal__doc__),
    GETTER_EXCEPTION("OperationalError", Cego_OperationalError, exception_operational__doc__),
    GETTER_EXCEPTION("PoolError", Cego_PoolError, exception_pool__doc__),
    GETTER_EXCEPTION("DataError", Cego_DataError, exception_data__doc__),
    {NULL}
};

static PyMethodDef
CegoConnection_Methods[] =
{
    /* PEP-249 methods */
    {"close", (PyCFunction)CegoConnection_close,
     METH_NOARGS,
     connection_close__doc__
    },
    {"connect", (PyCFunction)CegoConnection_connect,
     METH_VARARGS | METH_KEYWORDS,
     connection_connect__doc__
    },
    /* additional methods */
    
    { "_execute_command", 
      (PyCFunction)CegoConnection_executecommand,
      METH_O,
      "For internal use only"
    },

    { "ping",
      (PyCFunction)CegoConnection_ping,
      METH_NOARGS,
      connection_ping__doc__
    },
    
    {"_cegodb_get_info", (PyCFunction)CegoConnection_getinfo,
     METH_O,
     "For internal use only"},
    
    {NULL} /* always last */
};

static struct
PyMemberDef CegoConnection_Members[] =
{
    {"dsn",
        T_OBJECT,
        offsetof(CegoConnection, dsn),
        READONLY,
        "Data source name (dsn)"},
    {"_closed",
        T_BOOL,
     offsetof(CegoConnection, closed),
        READONLY,
        "Indicates if connection was closed"},
    {NULL} /* always last */
};

static int CegoConnection_Initialize(CegoConnection *self,
			  PyObject *args,
			  PyObject *dsnargs)
{
    char *dsn=NULL, *host=NULL, *user=NULL, *password=NULL, *tableset=NULL, *protocol=NULL;

    // we disable logfile
    char *logfile= NULL;

    // printf("CegoConnection_Initialize called\n");
    
    unsigned int port= 0;
    
    char *pool_name= 0;
    uint32_t pool_size= 0;
    
    if (!PyArg_ParseTupleAndKeywords(args, dsnargs,
				     "|zzizzzzzi:connect",
				     dsn_keys,
				     &dsn, &host, &port, &protocol, &tableset, &user, &password,
				     &pool_name, &pool_size))
    {
        return -1;
    }

    if (dsn)
    {
        cegoutil_throw_exception(NULL, Cego_ProgrammingError, 1,
                "dsn keyword is not supported");
        return -1;
    }

    if ( (self->cgdb = cego_connect(host, port, protocol, tableset, user, password, logfile)) == NULL )
    {
        cegoutil_throw_exception(NULL, Cego_DatabaseError, 1,
                cgerrmsg);
        return -1;
    }

    
    if (PyErr_Occurred())
        return -1;

    self->closed=0;
    
    return 0;
}

static int CegoConnection_traverse(
        CegoConnection *self,
        visitproc visit,
        void *arg)
{
    return 0;
}

static PyObject *CegoConnection_repr(CegoConnection *self)
{
    char cobj_repr[384];
    snprintf(cobj_repr, 384, "<cego.connection at %p>",
	     self);
    return PyUnicode_FromString(cobj_repr);
}

PyTypeObject CegoConnection_Type = {
    PyVarObject_HEAD_INIT(NULL, 0)
    .tp_name = "cegodb.connection",
    .tp_basicsize = (Py_ssize_t)sizeof(CegoConnection),
    .tp_repr = (reprfunc)CegoConnection_repr,
    .tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    .tp_doc = connection__doc__,
    .tp_new = PyType_GenericNew,
    .tp_traverse = (traverseproc)CegoConnection_traverse,
    .tp_methods = (struct PyMethodDef *)CegoConnection_Methods,
    .tp_members = (struct PyMemberDef *)CegoConnection_Members,
    .tp_getset = CegoConnection_sets,
    .tp_init = (initproc)CegoConnection_Initialize,
    .tp_alloc = PyType_GenericAlloc,
    .tp_finalize = (destructor)CegoConnection_finalize
};

PyObject* CegoConnection_connect(
        PyObject *self,
        PyObject *args,
        PyObject *kwargs)
{
    CegoConnection *c;

    // printf("CegoConnection_connect called\n");
    
    if ( ! ( c= (CegoConnection *)PyType_GenericAlloc(&CegoConnection_Type, 1) ) )
        return NULL;

    if (CegoConnection_Initialize(c, args, kwargs))
    {
        Py_DECREF(c);
        return NULL;
    }
    
    return (PyObject *) c;
}

static
void CegoConnection_finalize(CegoConnection *self)
{
    if (self)
    {
        if (self->cgdb)
        {
	    cego_disconnect(self->cgdb);
            self->cgdb= NULL;
        }
    }
}


static PyObject* CegoConnection_executecommand(CegoConnection *self,
                             PyObject *command)
{
    const char *cmd;
    
    cmd= PyUnicode_AsUTF8AndSize(command, NULL);
    
    if ( cego_query(self->cgdb, cmd, 0) != 0 )
    {
	// printf("Cego Error : %s\n", cgerrmsg);

	cegoutil_throw_exception(self->cgdb, Cego_ProgrammingError, 0,
			     cgerrmsg);	       	
	return NULL;
    }
    
    Py_RETURN_NONE;
}

static PyObject* CegoConnection_close(CegoConnection *self)
{
    // CEGO_CHECK_CONNECTION(self, NULL);
    
    /* Todo: check if all the cursor stuff is deleted (when using prepared
       statements this should be handled in mysql_close) */
    if ( self->cgdb )
    {
	cego_disconnect(self->cgdb);
	self->cgdb= NULL;
	self->closed= 1;
    }
    Py_RETURN_NONE;
}

static PyObject* CegoConnection_exception(PyObject *self, void *closure)
{
    PyObject *exception = *(PyObject **)closure;

    Py_INCREF(exception);
    return exception;
}

static PyObject* CegoConnection_ping(CegoConnection *self)
{
    if ( cego_query (self->cgdb, "uptime", 0) != 0 )
    {
        cegoutil_throw_exception(self->cgdb, Cego_InterfaceError, 0, NULL);
        return NULL;
    }

    Py_RETURN_NONE;
}

static PyObject* CegoConnection_getinfo(CegoConnection *self, PyObject *optionval)
{
    char* opt = cegoutil_object2string(optionval);   

    // first handle integer values
	   
    if ( strcmp(opt, "port") == 0 )
	return PyLong_FromLong((long)cego_port(self->cgdb));
        
    // now handle string values
    
    char *pVal = 0;
    
    if ( strcmp(opt, "dbname") == 0 )
	pVal = cego_dbname(self->cgdb);
    else if ( strcmp(opt, "version") == 0 )      
	pVal = cego_version(self->cgdb);
    else if ( strcmp(opt, "tableset") == 0 )      
	pVal = cego_tableset(self->cgdb);
    else if ( strcmp(opt, "user") == 0 )      
	pVal = cego_user(self->cgdb);
    else if ( strcmp(opt, "host") == 0 )      
	pVal = cego_server(self->cgdb);
    else if ( strcmp(opt, "tableset") == 0 )      
	pVal = cego_tableset(self->cgdb);
    
    if ( pVal )
    {
	PyObject* pValue = PyUnicode_FromString(pVal);
	return pValue;
    }
    Py_RETURN_NONE;
}

static PyObject *CegoConnection_connection_id(CegoConnection *self)
{
    return PyLong_FromUnsignedLong((unsigned long)0);
}


