 /*****************************************************************************
  Copyright (C) 2024 by Björn Lemke

  This code is derived from MariaDB/Connector written by Georg Richter

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not see <http://www.gnu.org/licenses>
  or write to the Free Software Foundation, Inc.,
  51 Franklin St., Fifth Floor, Boston, MA 02110, USA
 ****************************************************************************/

#include <cego_python.h>
#include <datetime.h>

static void
CegoCursor_finalize(CegoCursor *self);

static PyObject *
CegoCursor_close(CegoCursor *self);

static PyObject *
CegoCursor_prepareQuery(CegoCursor *self);

static PyObject *
CegoCursor_bindQuery(CegoCursor *self);

static PyObject *
CegoCursor_executeQuery(CegoCursor *self);

static PyObject *
CegoCursor_executeProc(CegoCursor *self);

static PyObject *
CegoCursor_InitResultSet(CegoCursor *self);

static PyObject *
CegoCursor_description(CegoCursor *self);

static PyObject *
CegoCursor_fetchone(CegoCursor *self);

static PyObject *
CegoCursor_fetchrows(CegoCursor *self, PyObject *rows);

static PyObject *Cego_row_count(CegoCursor *self);
static PyObject *Cego_row_number(CegoCursor *self);
static PyObject *CegoCursor_warnings(CegoCursor *self);
static PyObject *CegoCursor_closed(CegoCursor *self);

static PyGetSetDef CegoCursor_sets[]=
{
    {"description", (getter)CegoCursor_description, NULL,
        cursor_description__doc__, NULL},
    {"rowcount", (getter)Cego_row_count, NULL,
        NULL, NULL},
    {"warnings", (getter)CegoCursor_warnings, NULL,
        cursor_warnings__doc__, NULL},
    {"closed", (getter)CegoCursor_closed, NULL,
        cursor_closed__doc__, NULL},
    {"rownumber", (getter)Cego_row_number, NULL,
        cursor_rownumber__doc__, NULL},

    {NULL}
};

static PyMethodDef CegoCursor_Methods[] =
{
    /* PEP-249 methods */
    {"close", (PyCFunction)CegoCursor_close,
     METH_NOARGS,
     cursor_close__doc__},
    {"fetchone", (PyCFunction)CegoCursor_fetchone,
     METH_NOARGS,
     cursor_fetchone__doc__,},
    {"fetchrows", (PyCFunction)CegoCursor_fetchrows,
     METH_O,
     NULL},
    {"next", (PyCFunction)CegoCursor_fetchone,
     METH_NOARGS,
     cursor_next__doc__},

    /* internal helper functions */
    {"_preparequery", (PyCFunction)CegoCursor_prepareQuery,
     METH_NOARGS,
     NULL},
    {"_bindquery", (PyCFunction)CegoCursor_bindQuery,
     METH_NOARGS,
     NULL},
    {"_executequery", (PyCFunction)CegoCursor_executeQuery,
     METH_NOARGS,
     NULL},
    {"_executeproc", (PyCFunction)CegoCursor_executeProc,
     METH_NOARGS,
     NULL},
    {"_initresult", (PyCFunction)CegoCursor_InitResultSet,
     METH_NOARGS,
     NULL},
    {NULL} /* always last */
};

static struct PyMemberDef CegoCursor_Members[] =
{
    {"_statement",
        T_OBJECT,
        offsetof(CegoCursor, statement),
        0,
        cursor_statement__doc__},
    {"_numparam",
        T_UINT,
        offsetof(CegoCursor, numparam),
        0,
        MISSING_DOC},
    {"_data",
        T_OBJECT,
        offsetof(CegoCursor, data),
        0,
        MISSING_DOC},
    {"_spname",
        T_OBJECT,
        offsetof(CegoCursor, spname),
        0,
        MISSING_DOC},
    {"field_count",
        T_UINT,
        offsetof(CegoCursor, field_count),
        READONLY,
        cursor_field_count__doc__},
    {"_rownumber",
     T_ULONGLONG,
     offsetof(CegoCursor, row_number),
     0,
     NULL},
    {"_outparam",
     T_OBJECT,
     offsetof(CegoCursor, outparam),
     0,
     NULL},

    {NULL}
};

static int CegoCursor_initialize(CegoCursor *self, PyObject *args,
        PyObject *kwargs)
{
    char *key_words[]= {"", "reserved1", "reserved2", "reserved3", NULL};
    PyObject *connection;
    unsigned long reserved1= 0,
	reserved2= 0,
	reserved3= 0;

    if (!self)
        return -1;

    /*
      
      Notes for args handling :

      O! (object) [typeobject, PyObject *]

      Store a Python object in a C object pointer. 
      This is similar to O, but takes two C arguments: the first is the address of a Python type object,
      the second is the address of the C variable (of type PyObject*) into which the object pointer is stored. 
      If the Python object does not have the required type, TypeError is raised.

      This is, how we get connection handle ( CegoConnection defined in cego_python.h ) from python connection type object 
      ( CegoConnection_Type declared in cego_connection.c )
    */
    
    if (!PyArg_ParseTupleAndKeywords(args, kwargs,
                "O!|kkk", key_words, &CegoConnection_Type, &connection,
                &reserved1, &reserved2, &reserved3))
        return -1;

    if (! ((CegoConnection*)connection)->cgdb)
    {
        cegoutil_throw_exception(NULL, Cego_ProgrammingError, 0,
				 "Connection isn't valid anymore");
        return -1;
    }
    
    self->connection= (CegoConnection *)connection;
    self->stmtHandle= NULL;
    self->fetchHandle= NULL;
    self->values = NULL;
    self->outparam = NULL;
    self->blobCount = 0;
    self->outParamCount = 0;
    
    return 0;
}

// dummy function required from PyTypeObject

static int CegoCursor_traverse(
        CegoCursor *self,
        visitproc visit,
        void *arg)
{
    return 0;
}

static PyObject *CegoCursor_repr(CegoCursor *self)
{
    char cobj_repr[384];

    if (!self->closed)
        snprintf(cobj_repr, 384, "<cego.cursor at %p>", self);
    else
        snprintf(cobj_repr, 384, "<cego.cursor (closed) at %p>",
                self);
    return PyUnicode_FromString(cobj_repr);
}

PyTypeObject CegoCursor_Type =
{
    PyVarObject_HEAD_INIT(NULL, 0)
    .tp_name = "cego.cursor",
    .tp_basicsize= (Py_ssize_t)sizeof(CegoCursor),
    .tp_repr= (reprfunc)CegoCursor_repr, 
    .tp_flags= Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    .tp_doc= "Returns a cego cursor object", 
    .tp_traverse= (traverseproc)CegoCursor_traverse,
    .tp_methods= (struct PyMethodDef *)CegoCursor_Methods,
    .tp_members= (struct PyMemberDef *)CegoCursor_Members,
    .tp_getset= CegoCursor_sets,
    .tp_init= (initproc)CegoCursor_initialize,
    .tp_new= PyType_GenericNew,
    .tp_finalize= (destructor)CegoCursor_finalize
};

static void cg_cursor_close(CegoCursor *self)
{
    if ( !self->closed )
    {
        if (self->stmtHandle)
        {
	    cego_free_stmt(self->stmtHandle);
            self->stmtHandle= NULL;
        }

	self->fields= NULL;

	if (self->values)
	{
	    PyMem_RawFree(self->values);
	    self->values= NULL;
	}
	
        self->closed= 1;
    }
}

static PyObject* CegoCursor_close(CegoCursor *self)
{
    cg_cursor_close(self);
    Py_RETURN_NONE;
}

static void CegoCursor_finalize(CegoCursor *self)
{
    if (self->connection && self->connection->cgdb)
        cg_cursor_close(self);
}

static int Cego_GetFieldInfo(CegoCursor *self)
{
    self->row_number= 0;

    if (self->field_count)
    {
    }
    return 0;
}

PyObject *CegoCursor_InitResultSet(CegoCursor *self)
{

    if (self->values)
    {
	PyMem_RawFree(self->values);
	self->values= NULL;
    }

    if (self->field_count)
    {
        if (Cego_GetFieldInfo(self))
        {
            return NULL;
        }
	
        if (!(self->values= (PyObject**)PyMem_RawCalloc(self->field_count, sizeof(PyObject *))))
            return NULL;
    }

    Py_RETURN_NONE;
}

// actually not supported
static PyObject *CegoCursor_description(CegoCursor *self)
{
    Py_RETURN_NONE;
}

static PyObject *
CegoCursor_warnings(CegoCursor *self)
{
    return PyLong_FromLong((long)0);
}

static PyObject
*CegoCursor_closed(CegoCursor *self)
{
    if (self->closed || self->connection->cgdb == NULL)
        Py_RETURN_TRUE;
    Py_RETURN_FALSE;
}

static PyObject *
CegoCursor_prepareQuery(CegoCursor *self)
{
    // PyObject_Print(self->statement, stdout, 0);
    // printf("\n");

    // PyObject* str = PyUnicode_AsEncodedString(self->statement, "utf-8", "strict");
    // char *s = PyBytes_AS_STRING(str);

    char* s = cegoutil_object2string(self->statement);
    
    // printf("Preparing query <%s>\n", s);
    
    self->stmtHandle = cego_prepare(s);

    self->row_count = 0;
    
    Py_RETURN_TRUE;
    
}

static PyObject *
CegoCursor_bindQuery(CegoCursor *self)
{
    
    const char* tp = Py_TYPE(self->data)->tp_name;

    if ( strcmp(tp, "tuple") == 0 )
    {
	for (int i=0; i < self->numparam; i++)
	{
	    PyObject *obj = PyTuple_GetItem(self->data, i);
	    cegoutil_bind_internal(self, obj, i+1, 0);	 	    
	}
    }
    else
    {
	cegoutil_bind_internal(self, self->data, 1, 0);
    }

    if (PyErr_Occurred())
    {
        return NULL;
    }
	
    
    Py_RETURN_TRUE;

}


static PyObject *
CegoCursor_executeQuery(CegoCursor *self)
{
    
    // printf("Creating fetch handle ...\n");
    self->fetchHandle = cego_allocate_fetch();
    

    if ( self->connection->cgdb )
    {
	if ( cego_execute(self->connection->cgdb, self->stmtHandle, self->fetchHandle) != 0)
	{
	    printf("Free blobs ...\n");
	    cegoutil_free_blob(self);
	    
	    printf("Error : %s\n", cgerrmsg);
	    
	    cegoutil_throw_exception(NULL, Cego_ProgrammingError, 0,
                cgerrmsg);
	
	    
	    return NULL;
	}
    }
    else
    {
	cegoutil_throw_exception(NULL, Cego_InterfaceError, 0,
			     "Invalid db handle");
	return NULL;
    }

    // if no result sets available we set rowcount to affected
    if ( cego_isfetchable(self->fetchHandle) == 0 )
    {
	self->field_count = 0;
	self->row_count += cego_getaffected(self->connection->cgdb );
	// printf("Setting row count to %d\n", self->row_count);
    }
    else
    {
	self->field_count = cego_num_col(self->fetchHandle);
    }

    self->closed=0;
    
    Py_RETURN_NONE;
    
}


static PyObject *
CegoCursor_executeProc(CegoCursor *self)
{    
    if ( self->connection->cgdb )
    {
	
	char* spname = cegoutil_object2string(self->spname);
	char* statement = cegoutil_object2string(self->statement);
	
	// printf("Preparing query <%s>\n", s);
	
	self->stmtHandle = cego_prepare(statement);
	
	self->row_count = 0;
	
	char inout[100];
	cegoutil_getprocinout(self, spname, inout);
	// printf("IN Out Info = <%s>\n", inout);

	const char* tp = Py_TYPE(self->data)->tp_name;
	
	if ( strcmp(tp, "tuple") == 0 )
	{
	    int i=0;
	    while ( inout[i] != 0 )
	    {		
		PyObject *obj = PyTuple_GetItem(self->data, i);
				
		if ( inout[i] == 'i' )
		{
		    cegoutil_bind_internal(self, obj, i+1, 0);
		}
		else
		    cegoutil_bind_internal(self, obj, i+1, 1);
		i++;
	    }	    
	}
	else
	{	    
	    if ( inout[0] == 'i' )
	    {
		cegoutil_bind_internal(self, self->data, 1, 0);
	    }
	    else
	    {
		cegoutil_bind_internal(self, self->data, 1, 1);
	    }
	}
	          
	// printf("executing stmt  <%s>\n", statement);
       
	if ( cego_execute(self->connection->cgdb, self->stmtHandle, 0) != 0)
	{	    
	    // printf("Error : %s\n", cgerrmsg);
	    
	    cegoutil_throw_exception(NULL, Cego_ProgrammingError, 0,
                cgerrmsg);
		    
	    return NULL;
	}


	if ( ! (self->outparam = PyList_New(0)))
	{
	    return NULL;
	}
		
	int i=0;
	int outCount=0;
	while ( inout[i] != 0 )
	{
	    // PyObject *obj = PyTuple_GetItem(self->data, i);
	    if ( inout[i] == 'o' )
	    {

		int len = self->allocatedOutParam[outCount]->len;
		int type= self->allocatedOutParam[outCount]->type;
		char* val= self->allocatedOutParam[outCount]->val;

		PyObject* obj = cegoutil_cgval2object(type, len, val);


		// printf("XXXXX FROM cgval2objec : ");
		// PyObject_Print(obj, stdout, 0);
		
		
		// PyObject* obj = PyUnicode_FromStringAndSize((const char *)"Hello world",
		//  			(Py_ssize_t)12);

		
		PyList_Append(self->outparam, obj);
		/* CONPY-99: Decrement Row to prevent memory leak */
		Py_DECREF(obj);
		
		// cegoutil_assign_value(obj, self->allocatedOutParam[outCount]);
		outCount++;
	    }
	    i++;
	}
    }
    else
    {
	cegoutil_throw_exception(NULL, Cego_InterfaceError, 0,
			     "Invalid db handle");
	return NULL;
    }
    
    Py_RETURN_NONE;
    
}


static PyObject *
CegoCursor_fetchone(CegoCursor *self)
{
    PyObject *row;
    int field_count= self->field_count;

    // printf("C fetchone called\n");
    
    if (PyErr_Occurred())
    {
        return NULL;
    }

    self->fields = (CGVal*)malloc(sizeof(CGVal)*field_count);
    
    for ( int i=0; i<field_count; i++ )
	self->fields[i].val=0;

    
    int colnum = 0;
    if ( self->fetchHandle )
	colnum = cego_fetch (self->fetchHandle, self->fields, field_count);
    else
    {
	cegoutil_throw_exception(NULL, Cego_DatabaseError, 0,
				 "No valid fetch handle");
        return NULL;
    }
    
    if ( colnum > 0 ) 
    {	
	int i=0;	
	while ( i < colnum )
	{
	    cegoutil_field_fetch(self, self->fields[i].val, i);	    
	    i++;
	}
    }
    else
    {
	Py_RETURN_NONE;
    }
    
    row = PyTuple_New(self->field_count);

    // printf("Setting values\n");
    for (int i= 0; i < field_count; i++)
    {	
	PyTuple_SET_ITEM(row, i, (self)->values[i]);
    }
    self->row_number++;
    
    return row;

}

static PyObject *
CegoCursor_fetchrows(CegoCursor *self, PyObject *rows)
{
    PyObject *rowList;
    unsigned int field_count= self->field_count;
    long long row_count;


    if (!field_count)
    {
        cegoutil_throw_exception(NULL, Cego_ProgrammingError, 0,
                "Cursor doesn't have a result set");
        return NULL;
    }

    if (!CHECK_TYPE_NO_NONE(rows, &PyLong_Type)) {
        PyErr_SetString(PyExc_TypeError, "Parameter must be an integer value");
        return NULL;
    }

    row_count= PyLong_AsLongLong(rows);

    if ( ! (rowList = PyList_New(0)))
    {
        return NULL;
    }
    
    long long i=0;
    while ( ( row_count == -1 || i < row_count ) && cego_isfetchable(self->fetchHandle))
    {	
        PyObject *row;

        self->row_number++;
	
        if ( ( row= CegoCursor_fetchone(self)) != Py_None )
        {	   
	    PyList_Append(rowList, row);
	    /* CONPY-99: Decrement Row to prevent memory leak */
	    Py_DECREF(row);
	}
	i++;
    }

    self->row_count=i;
    return rowList;
}


static PyObject *
Cego_row_count(CegoCursor *self)
{
    // printf("Cursor Rowcount = %d\n", self->row_count);
    return PyLong_FromLongLong(self->row_count);
}

static PyObject *
Cego_row_number(CegoCursor *self)
{
    if (!self->field_count) {
        Py_RETURN_NONE;
    }
    return PyLong_FromLongLong(self->row_number);
}
