 /*****************************************************************************
  Copyright (C) 2024 by Björn Lemke

  This code is derived from MariaDB/Connector written by Georg Richter

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not see <http://www.gnu.org/licenses>
  or write to the Free Software Foundation, Inc.,
  51 Franklin St., Fifth Floor, Boston, MA 02110, USA
 ****************************************************************************/

#include <cego_python.h>
#include <datetime.h>

/* Exceptions */
PyObject *Cego_InterfaceError;
PyObject *Cego_Error;
PyObject *Cego_DatabaseError;
PyObject *Cego_DataError;
PyObject *Cego_PoolError;
PyObject *Cego_OperationalError;
PyObject *Cego_IntegrityError;
PyObject *Cego_InternalError;
PyObject *Cego_ProgrammingError;
PyObject *Cego_NotSupportedError;
PyObject *Cego_Warning;


int cegoutil_datetime_init(void)
{
    PyDateTime_IMPORT;

    if (!PyDateTimeAPI) {
        PyErr_SetString(PyExc_ImportError, "DateTimeAPI initialization failed");
        return 1;
    }
    return 0;
}

/**
 cego_throw_exception()
 @brief  raises an exception

 @param  handle[in]            a connection or statement handle
 @param  exception_type[in]    type of exception
 @param  handle_type[in]       -1 no handle (use error_no)
                                0 cgdb
                                1 pStmt
 @param  message[in]           Error message. If message is NULL, the error
                               message will be retrieved from specified handle.
 @param  ... [in]              message parameter

 @return void

*/
void cegoutil_throw_exception(CGDB* cgdb,
			  PyObject *exception_type,
			  int8_t is_statement,
			  const char *message,
			  ...)
{
    va_list ap;
    PyObject *ErrorMsg= 0;
    PyObject *ErrorNo= 0;
    PyObject *SqlState= 0;
    PyObject *Exception= 0;
    
    if (message)
    {
	// ErrorNo= PyLong_FromLong(CR_UNKNOWN_ERROR);
	SqlState= PyUnicode_FromString("HY000");
	va_start(ap, message);
	ErrorMsg= PyUnicode_FromFormatV(message, ap);
	va_end(ap);
    }
    
    
    if (!(Exception= PyObject_CallFunctionObjArgs(exception_type, ErrorMsg, NULL)))
    {
	PyErr_SetString(PyExc_RuntimeError,
			"Failed to create exception");
	return;
    }
    
    PyObject_SetAttr(Exception, PyUnicode_FromString("sqlstate"), SqlState);
    PyObject_SetAttr(Exception, PyUnicode_FromString("errno"), ErrorNo);
    PyObject_SetAttr(Exception, PyUnicode_FromString("errmsg"), ErrorMsg);
    /* For MySQL Connector/Python compatibility */
    PyObject_SetAttr(Exception, PyUnicode_FromString("msg"), ErrorMsg);
    PyErr_SetObject(exception_type, Exception);
    Py_XDECREF(ErrorMsg);
    Py_XDECREF(ErrorNo);
    Py_XDECREF(SqlState);
}

int
cegoutil_bind_internal(CegoCursor *self, PyObject* obj, int pos, int inout)
{
    
    const char* p = Py_TYPE(obj)->tp_name;
    
    CGVal *pVal = 0;
    if ( strcmp(p, "int") == 0 )
    {
	int i = (int) PyLong_AsLong(obj);
	pVal = cego_allocate_value(CG_INT, sizeof(int), &i);	    
    }
    else if ( strcmp(p, "str") == 0 )
    {
	char* s = cegoutil_object2string(obj);
	pVal = cego_allocate_value(CG_VARCHAR, strlen(s), s);	
    }
    else if ( strcmp(p, "float") == 0 )
    {
	double d = PyFloat_AsDouble(obj);
	pVal = cego_allocate_value(CG_DOUBLE, sizeof(double), &d);
    }
    else if ( strcmp(p, "bytes") == 0 )
    {

	// printf("Detected blob handling\n");
	
	CGBlob b;
	b.buf = (unsigned char*)PyBytes_AsString(obj);       
	b.len = PyBytes_Size(obj);
	b.pageId = 0;

	// printf("Blob size is %d\n", b.len);
	
	cego_putblob (self->connection->cgdb, &b);
	// printf("Created blob with ref : [%llu]\n", b.pageId); 

	self->allocatedBlob[self->blobCount]=b.pageId;
	self->blobCount++;
	
	pVal = cego_allocate_value(CG_BLOB, sizeof(PageIdType), &b.pageId);	
    }
    else
    {
	PyErr_SetString(PyExc_ImportError, "Unknown datatype");
	return -1;    
    }
        
    if ( pVal )
    {
	if ( inout == 0 )
	    cego_bind_in(self->stmtHandle, pVal, pos);
	else
	{
	    cego_bind_out(self->stmtHandle, pVal, pos);
	    self->allocatedOutParam[self->outParamCount] = pVal;
	    self->outParamCount++;
	}
    }
}

void cegoutil_free_blob(CegoCursor* self)
{
    for ( int i=0; i<self->blobCount; i++ )
    {
    	CGBlob b;
	b.pageId = self->allocatedBlob[i];

	printf("Delete blob with page %d\n", b.pageId);
	
	cego_delblob (self->connection->cgdb, &b);
    }
}

void
cegoutil_field_fetch(CegoCursor *self, char *data, unsigned int column)
{

    int len = self->fields[column].len;
    int type= self->fields[column].type;

    if (!data)
      type= CG_NULL;


    self->values[column] = cegoutil_cgval2object(type, len, data);
 
} 

PyObject* cegoutil_cgval2object(int type, int len, char *data)
{

    // printf("Type = %d\n", type);
    // printf("Length = %d\n", len);

    switch (type)
    {
    case CG_NULL:
    {
	Py_INCREF(Py_None);
	return Py_None;
    }
    case CG_TINYINT:
    {
        char* pTI = data;
	return  PyLong_FromLong((char)*pTI);
    }
    case CG_SMALLINT:
    {
	short *pSI = data;
	return  PyLong_FromLong((short)*pSI);
    }
    case CG_INT:
    {
	int *pI = data;
	return  PyLong_FromLong((int)*pI);
    }
    case CG_LONG:
    {
	long *pLong = data;
	return  PyLong_FromLong((long)*pLong);
    }
    case CG_FLOAT:
    {
	float* pFloat = data;
	return PyFloat_FromDouble((float)*pFloat);
    }	
    case CG_DOUBLE: 
    {
	double* pDouble = data;
	return PyFloat_FromDouble((double)*pDouble);
    }
    case CG_DATETIME:
    {
	// printf("Data is %d\n", data);
	unsigned long long *pL = data;

	struct tm *pTM = localtime( pL );
	// printf("got local time\n");

	int second = pTM->tm_sec;
	int minute = pTM->tm_min;
	int hour = pTM->tm_hour;
	int month = pTM->tm_mon + 1;
	int day = pTM->tm_mday;
	int year = pTM->tm_year + 1900;
	    
	// Please note : For the following call, PyDateTime_IMPORT macro has to be called
	// Otherwise we get a segmentation fault
	return PyDateTime_FromDateAndTime(year,month,day,hour,minute,second,0);
    }
    case CG_BLOB:
    case CG_CLOB:
    {
	/*
	if (length[column] > self->fields[column].max_length)
	{
	    self->fields[column].max_length= length[column];
	}
	if (self->fields[column].charsetnr== CHARSET_BINARY)
	{
	    self->values[column]= 
		PyBytes_FromStringAndSize((const char *)data,
					  (Py_ssize_t)length[column]);
	}
	else {
	    self->values[column]= 
		PyUnicode_FromStringAndSize((const char *)data, 
					    (Py_ssize_t)length[column]);
	}
	*/
	break;
    }
    case CG_VARCHAR:
    case CG_DECIMAL:
    case CG_BIGINT:
    {
	// printf("Check value %s:\n", (char*)data);	
	return PyUnicode_FromStringAndSize((const char *)data,
					(Py_ssize_t)len);
    }
    default:
	break;
    }

    return Py_None;
    
}

char* cegoutil_object2string(PyObject *obj)
{
    PyObject* str = PyUnicode_AsEncodedString(obj, "utf-8", "strict");
    char *s = PyBytes_AS_STRING(str);
    return s;
}

/*
  The following utility function retrieves information about procedure IO-Parameters
  This is neede, to bind procedure parameter with appropriate bind function ( cego_bind_in / cego_bind_out ) 
*/

void  cegoutil_getprocinout(CegoCursor* self, char* procName, char* inoutinfo)
{

    char stmt[100];
    sprintf(stmt, "desc procedure %s;", procName);
    
    CGDB *cgdb = self->connection->cgdb;
    
    // printf("Creating fetch handle ...\n");
    CGFetch* cgfetch = cego_allocate_fetch();

    // printf("Making query ...\n");
    if ( cego_query(cgdb, stmt, cgfetch) != 0 )
    {
	cego_free_fetch(cgfetch);    
	printf("Error : %s\n", cgerrmsg);
	return;
    }
    
    // printf("Size=%d\n", cego_num_col(cgfetch));
    
    CGVal cgval[5];
    cgval[0].val=0; // ATTR
    cgval[1].val=0; // TYPE
    cgval[2].val=0; // LENGTH
    cgval[3].val=0; // DIM
    cgval[4].val=0; // INOUT
    
    
    int colnum;
    int i=0;
    while ( ( colnum = cego_fetch (cgfetch, cgval, 5)) > 0 )
    {
	// we just need to get information of col 5 
	if ( strcmp( (char*)cgval[4].val, "in") == 0 )
	    inoutinfo[i]='i';
	if ( strcmp( (char*)cgval[4].val, "out") == 0 )
	    inoutinfo[i]='o';
	i++;
    }
    inoutinfo[i]=0;
    
    // printf("Fetch ok\n");

    cego_free_fetch(cgfetch);    
    // cego_disconnect(cgdb);
}
