/******************************************************************************
  Copyright (C) 2024 by Björn Lemke

  This code is derived from MariaDB/Connector written by Georg Richter

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Library General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Library General Public License for more details.

  You should have received a copy of the GNU Library General Public
  License along with this library; if not see <http://www.gnu.org/licenses>
  or write to the Free Software Foundation, Inc.,
  51 Franklin St., Fifth Floor, Boston, MA 02110, USA
 ******************************************************************************/

#include "cego_python.h"
#include <structmember.h>
#include <datetime.h>

extern PyObject* CegoConnection_connect( PyObject *self,PyObject *args, PyObject *kwargs);

static PyMethodDef
Cego_Methods[] =
{
    /* PEP-249: mandatory */
    {"connect", (PyCFunction)CegoConnection_connect,
        METH_VARARGS | METH_KEYWORDS,
        module_connect__doc__},
    /* Todo: add methods for api functions which don't require
       a connection */
    {NULL} /* always last */
};

/* Cego module definition */
static struct PyModuleDef cego_module= {
    PyModuleDef_HEAD_INIT,
    "_cego",
    "Cego Connector for Python",
    -1,
    Cego_Methods
};

static int cego_datetime_init(void)
{
    PyDateTime_IMPORT;
    
    if (!PyDateTimeAPI) {
        PyErr_SetString(PyExc_ImportError, "DateTimeAPI initialization failed");
        return 1;
    }
    return 0;
}

static void cegodb_add_exception(PyObject *module,
        PyObject **exception,
        const char *exception_name,
        PyObject *base_exception,
        const char *doc,
        const char *object_name)
{
    *exception= PyErr_NewExceptionWithDoc(exception_name,
            doc,
            Cego_Error,
            NULL);

    Py_INCREF(*exception);
    PyModule_AddObject(module, object_name, *exception);
}


/*
  Cego module initialization function
  This function is called with module import
  Name must be PyInit__<packagename>
*/

PyMODINIT_FUNC PyInit__cegodb(void)
{
    PyObject *module= PyModule_Create(&cego_module);

    // printf("Pyinit__cegodb called\n");
	
    /* Initialize DateTimeAPI */
    if ( cego_datetime_init() || cegoutil_datetime_init() )
    {
        goto error;
    }

    /* setup connection type */
    
    Py_SET_TYPE(&CegoConnection_Type, &PyType_Type);
    if (PyType_Ready(&CegoConnection_Type) == -1)
    {
        goto error;
    }

    Py_INCREF(&CegoConnection_Type);
    PyModule_AddObject(module, "connection", (PyObject *)&CegoConnection_Type);

    /* setup cursor type */

    Py_SET_TYPE(&CegoCursor_Type, &PyType_Type);
    if (PyType_Ready(&CegoCursor_Type) == -1)
    {
        goto error;
    }
    Py_INCREF(&CegoCursor_Type);
    PyModule_AddObject(module, "cursor", (PyObject *)&CegoCursor_Type);


    /* setup error type */
    
    Cego_Error= PyErr_NewException("cegodb.Error",
            PyExc_Exception,
            NULL);
    Py_INCREF(Cego_Error);
    PyModule_AddObject(module, "Error", Cego_Error);
    
    cegodb_add_exception(module, &Cego_InterfaceError,
            "cegodb.InterfaceError", Cego_Error,
            exception_interface__doc__, "InterfaceError");
    cegodb_add_exception(module, &Cego_DatabaseError,
            "cegodb.DatabaseError", Cego_Error,
            exception_database__doc__, "DatabaseError");
    cegodb_add_exception(module, &Cego_OperationalError,
            "cegodb.OperationalError", Cego_Error,
            exception_operational__doc__, "OperationalError");
    cegodb_add_exception(module, &Cego_Warning,
            "cegodb.Warning", NULL, exception_warning__doc__, "Warning");
    cegodb_add_exception(module, &Cego_IntegrityError,
            "cegodb.IntegrityError", Cego_Error,
            exception_integrity__doc__, "IntegrityError");
    cegodb_add_exception(module, &Cego_InternalError,
            "cegodb.InternalError", Cego_Error,
            exception_internal__doc__, "InternalError");
    cegodb_add_exception(module, &Cego_ProgrammingError,
            "cegodb.ProgrammingError", Cego_Error,
            exception_programming__doc__, "ProgrammingError");
    cegodb_add_exception(module, &Cego_NotSupportedError,
            "cegodb.NotSupportedError", Cego_Error,
            exception_notsupported__doc__, "NotSupportedError");
    
    cegodb_add_exception(module, &Cego_DataError,
            "cegodb.DataError", Cego_DatabaseError,
            exception_data__doc__, "DataError");
    
    cegodb_add_exception(module, &Cego_PoolError,
            "cegodb.PoolError", Cego_Error,
            exception_pool__doc__, "PoolError");
    

    return module;
error:
    if ( PyErr_Occurred() )
    {
        return NULL;
    }
    PyErr_SetString(PyExc_ImportError, "cego module initialization failed.");
    return NULL;
}
